import numpy as np
from mdt import CompartmentTemplate, CompositeModelTemplate, FreeParameterTemplate, ProtocolParameterTemplate, \
    LibraryFunctionTemplate
from mdt.model_building.parameter_functions.transformations import ScaleTransform
from mdt.utils import voxelwise_vector_matrix_vector_product, create_covariance_matrix
from mot.library_functions import simpsons_rule


class LineshapeGaussian(LibraryFunctionTemplate):
    return_type = 'double'
    parameters = ['double f', 'double T2_b']
    cl_code = '''
        return sqrt(1.0/(2.0*M_PI)) * T2_b * exp(-pown(2.0 * M_PI * f * T2_b, 2) / 2.0);
    '''


class LineshapeLorentzian(LibraryFunctionTemplate):
    return_type = 'double'
    parameters = ['double f', 'double T2_b']
    cl_code = '''
        return (T2_b / M_PI) / (1.0 + pown(2.0 * M_PI * f * T2_b, 2));
    '''


class LineshapeSuperLorentzian(LibraryFunctionTemplate):
    return_type = 'double'
    parameters = ['double f', 'double T2_b']
    dependencies = ['superlorentzian_integrand', simpsons_rule('superlorentzian_integrand')]
    cl_code = '''
        superlorentzian_integrand_data data = {f, T2_b};
        return simpsons_rule_superlorentzian_integrand(0, 1, 50, &data);
    '''

    class superlorentzian_integrand(LibraryFunctionTemplate):
        return_type = 'double'
        parameters = ['double x', 'void* data']
        cl_code = '''
            double T2_b = ((superlorentzian_integrand_data*)data)->T2_b;
            double f = ((superlorentzian_integrand_data*)data)->f;

            double uterm = fabs(T2_b / (3.0 * x * x - 1.0));
            return sqrt(2.0/M_PI) * uterm * exp(-2.0 * pown(2.0 * M_PI * f * uterm, 2));
        '''
        cl_extra = '''
            typedef struct superlorentzian_integrand_data{
                double f; 
                double T2_b;
            } superlorentzian_integrand_data;
        '''


class LineshapeSuperLorentzianInterpolated(LibraryFunctionTemplate):
    """Interpolated super Lorentzian lineshape

    The required data for this lineshape was computed using the Python code::

        T2_b = 1e-5
        freq_count = 150
        freq_start = 500
        freq_stepsize = 500

        frequenties = np.linspace(freq_start, freq_start + freq_stepsize * (freq_count - 1), freq_count)

        grid = SuperLorentzian()().evaluate((frequenties, T2_b), freq_count)

    This can be used instead of the SuperLorentzian lineshape. If you need different interpolation steps, overwrite
    this template with a different interpolation matrix.
    """
    return_type = 'double'
    parameters = ['double f', 'double T2_b']
    dependencies = ['linear_cubic_interpolation']
    cl_code = '''
        double T2_b_scale = 1e-5;
        uint freq_count = 150;
        uint freq_start = 500;
        uint freq_stepsize = 500;

        double scale = T2_b / T2_b_scale;
        double sf = (f * scale - freq_start) / freq_stepsize;

        return scale * linear_cubic_interpolation(sf, freq_count, LineshapeSuperLorentzianInterpolated_lookup_table);
    '''
    cl_extra = '''
        constant float LineshapeSuperLorentzianInterpolated_lookup_table[] = {
             1.6351068497997384e-05, 1.3115492795811157e-05, 1.1241666681196103e-05, 9.869994321088873e-06, 
             8.79781041979886e-06, 7.914196403750282e-06, 7.159953545164688e-06, 6.501822309129753e-06, 
             5.9187370092896205e-06, 5.396337299169265e-06, 4.9244567517431185e-06, 4.495680550607755e-06, 
             4.104430808410283e-06, 3.7463832161993373e-06, 3.4180920723477155e-06, 3.1167423473158515e-06, 
             2.8399801093075133e-06, 2.5857933549322158e-06, 2.352426901287217e-06, 2.138321167528001e-06, 
             1.9420680293872054e-06, 1.7623789684262465e-06, 1.5980621239498786e-06, 1.448005833752447e-06, 
             1.3111669297816838e-06, 1.1865625196512177e-06, 1.0732643058325794e-06, 9.70394723648105e-07, 
             8.771243496989519e-07, 7.926701628937718e-07, 7.16294341733855e-07, 6.473033610647276e-07, 
             5.850472142475361e-07, 5.289186364606442e-07, 4.783522444252163e-07, 4.3282353928886254e-07, 
             3.918477441733863e-07, 3.549784671059634e-07, 3.2180619458758904e-07, 2.919566316414404e-07, 
             2.6508891145979857e-07, 2.4089370228913683e-07, 2.1909124145582068e-07, 1.9942932688388915e-07, 
             1.816812954884552e-07, 1.6564401579574187e-07, 1.5113591935325928e-07, 1.3799509221911866e-07, 
             1.2607744428542744e-07, 1.1525497058546612e-07, 1.0541411520910552e-07, 9.645424512226031e-08, 
             8.828623813911807e-08, 8.083118658736193e-08, 7.401921586922215e-08, 6.77884151670253e-08, 
             6.208387596636248e-08, 5.6856832855435643e-08, 5.2063900178291605e-08, 4.76663975386842e-08, 
             4.3629756832484325e-08, 3.992300339053272e-08, 3.651830390146855e-08, 3.3390574017895954e-08, 
             3.051713889451962e-08, 2.7877440332283366e-08, 2.545278468063438e-08, 2.322612615740714e-08, 
             2.1181880763075917e-08, 1.930576647769909e-08, 1.7584665922739697e-08, 1.6006508137241576e-08, 
             1.456016655247463e-08, 1.3235370647395598e-08, 1.2022629127292448e-08, 1.0913162789529515e-08, 
             9.898845524315148e-09, 8.972152146678366e-09, 8.126111970774798e-09, 7.35426722203145e-09, 
             6.650635539463799e-09, 6.009675952783081e-09, 5.426257829619494e-09, 4.895632380138091e-09, 
             4.413406382140566e-09, 3.975517851821481e-09, 3.5782134358287956e-09, 3.218027341065111e-09, 
             2.8917616514154244e-09, 2.5964679067291395e-09, 2.329429840138001e-09, 2.0881471861869747e-09, 
             1.8703204851526994e-09, 1.673836819037118e-09, 1.4967564226374533e-09, 1.3373001192848592e-09, 
             1.1938375356972415e-09, 1.0648760542152894e-09, 9.490504637290755e-10, 8.451132730484687e-10, 
             7.519256524742328e-10, 6.684489710045162e-10, 5.937368980523485e-10, 5.269280398213664e-10, 
             4.672390816396365e-10, 4.139584086223556e-10, 3.664401780502923e-10, 3.240988178310485e-10, 
             2.8640392636758843e-10, 2.5287555010143854e-10, 2.2307981593159215e-10, 1.9662489663708468e-10, 
             1.7315728835226004e-10, 1.5235838005851848e-10, 1.339412959635643e-10, 1.1764799253708605e-10, 
             1.0324659285825612e-10, 9.052894180324604e-11, 7.930836645783098e-11, 6.94176269789315e-11, 
             6.070704394759365e-11, 5.304278905255795e-11, 4.6305326716595796e-11, 4.038799502575617e-11, 
             3.51957150433901e-11, 3.064381828534206e-11, 2.6656982799280144e-11, 2.316826892927348e-11, 
             2.0118246456090952e-11, 1.7454205384183292e-11, 1.5129443197966466e-11, 1.3102621933071603e-11, 
             1.133718890302249e-11, 9.800855388841954e-12, 8.465128038970233e-12, 7.304888140311299e-12, 
             6.298014308985523e-12, 5.425044512317307e-12, 4.668873672633755e-12, 4.014483419548509e-12, 
             3.4487008515413167e-12, 2.9599834408260634e-12, 2.5382274687498563e-12, 2.174597613306697e-12, 
             1.861375526803235e-12, 1.5918254412904438e-12, 1.3600750230678129e-12, 1.1610098663476446e-12, 
             9.901801709822935e-13, 8.437182909232526e-13};
    '''


def produce_reduced_ramani_model(lineshape_method_name):
    """Create the model template of a reduced Ramani model which will use the given line shape.

    This produces composite models with the naming scheme "QMT_ReducedRamani_<lineshape_method_name>".

    Args:
        lineshape_method_name (str): the name of the lineshape method to use, one of
            Gaussian, Lorentzian, SuperLorentzian, GaussianInterpolated

    Returns:
        CompositeModelTemplate: the model template for the reduced Ramani model using the given line shape.
    """
    def extra_output_maps(results):
        np.warnings.filterwarnings('ignore')

        Rb, RM0a, fterm, gM0a, T1a_T2a, s0 = [
            np.squeeze(results[n]) for n in ('QMT.Rb', 'QMT.RM0a', 'QMT.fterm', 'QMT.gM0a', 'QMT.T1a_T2a', 'S0.s0')]
        Rb_std, RM0a_std, fterm_std, gM0a_std, T1a_T2a_std, s0_std = [
            np.squeeze(results.get(n + '.std', 0)) for n in ('QMT.Rb', 'QMT.RM0a', 'QMT.fterm',
                                                             'QMT.gM0a', 'QMT.T1a_T2a', 'S0.s0')]

        R_obs = 1. / results.get_input_data('T1')
        if len(R_obs.shape) > 1:
            R_obs = R_obs[:, 0]

        Ra = R_obs / (1.0 + ((RM0a * fterm * (Rb - R_obs)) / (Rb - R_obs + RM0a)))
        f = fterm * Ra / (1.0 + fterm * Ra)

        covars = create_covariance_matrix(
            f.shape[0], results, ['QMT.RM0a', 'QMT.fterm', 'QMT.Rb'], results.get('covariances', None))

        Ra_gradient = get_ra_gradient(R_obs, RM0a, fterm, Rb)
        Ra_std = np.nan_to_num(np.sqrt(voxelwise_vector_matrix_vector_product(Ra_gradient, covars, Ra_gradient)))

        f_gradient = get_f_gradient(R_obs, RM0a, fterm, Rb)
        f_std = np.nan_to_num(np.sqrt(voxelwise_vector_matrix_vector_product(f_gradient, covars, f_gradient)))

        values = {
            'QMT.PD': gM0a * s0,
            'QMT.T1_f': 1.0 / Ra,
            'QMT.T2_f': (1.0 / Ra) / T1a_T2a,
            'QMT.k_bf': RM0a * f / (1.0 - f),
            'QMT.f_b': f,
            'QMT.Ra': Ra,
        }

        values.update({
            'QMT.PD.std': np.nan_to_num(np.sqrt(gM0a**2 * s0_std + s0**2 * gM0a_std)),
            'QMT.T1_f.std': np.nan_to_num(np.sqrt(Ra_std/Ra**4)),
            'QMT.T2_f.std': np.nan_to_num(np.sqrt((Ra**2 * T1a_T2a_std + Ra_std * T1a_T2a**2)/(Ra**4 * T1a_T2a**4))),
            'QMT.k_bf.std': np.nan_to_num(np.sqrt((RM0a**2 * f_std + RM0a_std * f**2 * (f - 1.0)**2) / (f - 1.0)**4)),
            'QMT.f_b.std': f_std,
            'QMT.Ra.std': Ra_std,
        })
        return values

    def get_ra_gradient(R_obs, RM0a, fterm, Rb):
        """Compute the standard deviation of Ra with respect to RM0a, fterm, Rb"""
        return np.column_stack([
            -R_obs*fterm*(R_obs - Rb)**2/(RM0a*fterm*(R_obs - Rb) - 1.0*RM0a + 1.0*R_obs - 1.0*Rb)**2,
            RM0a*R_obs*(R_obs - Rb)*(RM0a - R_obs + Rb)/(RM0a*fterm*(R_obs - Rb) - 1.0*RM0a + 1.0*R_obs - 1.0*Rb)**2,
            -RM0a**2*R_obs*fterm/(RM0a*fterm*(R_obs - Rb) - 1.0*RM0a + 1.0*R_obs - 1.0*Rb)**2])

    def get_f_gradient(R_obs, RM0a, fterm, Rb):
        """Compute the standard deviation of f with respect to RM0a, fterm, Rb"""
        return np.column_stack([
            1.0 * R_obs * fterm**2 * (R_obs - Rb)**2 * (-RM0a * fterm * (R_obs - Rb) + RM0a - R_obs + Rb) / (
                    (RM0a * fterm * (R_obs - Rb) - 1.0 * RM0a + 1.0 * R_obs - 1.0 * Rb) * (
                        -1.0 * RM0a * fterm * (R_obs - Rb) + 1.0 * RM0a + R_obs * fterm * (
                            RM0a - R_obs + Rb) - 1.0 * R_obs + 1.0 * Rb)**2),
        R_obs * (RM0a - R_obs + Rb) * (-RM0a * fterm * (R_obs - Rb) * (
                    -1.0 * RM0a * fterm * (R_obs - Rb) + 1.0 * RM0a + R_obs * fterm * (
                        RM0a - R_obs + Rb) - 1.0 * R_obs + 1.0 * Rb) + 1.0 * R_obs * fterm * (
                                                   RM0a - R_obs + Rb)**2 + (RM0a * fterm * (
                    R_obs - Rb) - 1.0 * RM0a + 1.0 * R_obs - 1.0 * Rb) * (
                                                   -1.0 * RM0a * fterm * (R_obs - Rb) + 1.0 * RM0a + R_obs * fterm * (
                                                       RM0a - R_obs + Rb) - 1.0 * R_obs + 1.0 * Rb)) / (
                    (RM0a * fterm * (R_obs - Rb) - 1.0 * RM0a + 1.0 * R_obs - 1.0 * Rb) * (
                        -1.0 * RM0a * fterm * (R_obs - Rb) + 1.0 * RM0a + R_obs * fterm * (
                            RM0a - R_obs + Rb) - 1.0 * R_obs + 1.0 * Rb)**2),
        1.0 * RM0a**2 * R_obs * fterm**2 * (-RM0a * fterm * (R_obs - Rb) + RM0a - R_obs + Rb) / (
                    (RM0a * fterm * (R_obs - Rb) - 1.0 * RM0a + 1.0 * R_obs - 1.0 * Rb) * (
                        -1.0 * RM0a * fterm * (R_obs - Rb) + 1.0 * RM0a + R_obs * fterm * (
                            RM0a - R_obs + Rb) - 1.0 * R_obs + 1.0 * Rb)**2)])

    class QMT(CompositeModelTemplate):
        name = 'QMT_ReducedRamani_' + lineshape_method_name
        model_expression = 'S0 * QMT(QMT)'
        likelihood_function = 'Gaussian'
        fixes = {'QMT.Rb': 1}
        extra_optimization_maps = [extra_output_maps]

        class QMT(CompartmentTemplate):
            parameters = ('Rb', 'RM0a', 'fterm', 'T2_b', 'T1a_T2a', 'gM0a',
                          'f0', 'B1', 'TR', 'sat_f0', 'sat_angle', 'pulse_Trf', 'pulse_p1', 'pulse_p2')
            dependencies = ['Lineshape' + lineshape_method_name]
            cl_code = '''
                double lsv = Lineshape''' + lineshape_method_name + '''(sat_f0 + f0, T2_b);
                
                double w_cwpe = (B1 * sat_angle / pulse_p1) * sqrt(pulse_p2 / (pulse_Trf * TR));
                double R_rfb = M_PI * (w_cwpe * w_cwpe) * lsv;
        
                double S = gM0a * (Rb * RM0a * fterm + R_rfb + Rb + RM0a) /
                    ((RM0a * fterm) * (Rb + R_rfb) + (1.0 + pown(w_cwpe / 
                        (2 * M_PI * sat_f0), 2) * T1a_T2a) * (R_rfb + Rb + RM0a));
        
                return S;
            '''

            class Rb(FreeParameterTemplate):
                init_value = 1
                lower_bound = 0
                upper_bound = 10
                numdiff_info = {'use_upper_bound': False}

            class RM0a(FreeParameterTemplate):
                """R*M0a"""
                init_value = 10
                lower_bound = 2
                upper_bound = 100
                numdiff_info = {'scale_factor': 0.1, 'use_upper_bound': False, 'use_lower_bound': False}

            class fterm(FreeParameterTemplate):
                """f/(R_a*(1-f))"""
                init_value = 0.1
                lower_bound = 1e-12
                upper_bound = 1
                numdiff_info = {'use_upper_bound': False}

            class T2_b(FreeParameterTemplate):
                """T2_b"""
                init_value = 10e-6
                lower_bound = 1e-7
                upper_bound = 100e-6
                parameter_transform = ScaleTransform(1e6)
                numdiff_info = {'scale_factor': 1e5, 'use_upper_bound': False, 'use_lower_bound': False}

            class T1a_T2a(FreeParameterTemplate):
                """T1_a/T2_a"""
                init_value = 25
                lower_bound = 1
                upper_bound = 50
                numdiff_info = {'use_upper_bound': False}

            class gM0a(FreeParameterTemplate):
                """gM0_a"""
                init_value = 1
                lower_bound = 0.5
                upper_bound = 1.5
                numdiff_info = {'use_bounds': False}

            class f0(ProtocolParameterTemplate):
                pass

            class B1(ProtocolParameterTemplate):
                pass

            class TR(ProtocolParameterTemplate):
                pass

            class sat_f0(ProtocolParameterTemplate):
                pass

            class sat_angle(ProtocolParameterTemplate):
                pass

            class pulse_Trf(ProtocolParameterTemplate):
                pass

            class pulse_p1(ProtocolParameterTemplate):
                pass

            class pulse_p2(ProtocolParameterTemplate):
                pass


produce_reduced_ramani_model('Gaussian')
produce_reduced_ramani_model('Lorentzian')
produce_reduced_ramani_model('SuperLorentzian')
produce_reduced_ramani_model('SuperLorentzianInterpolated')



from mdt import CompartmentTemplate
import numpy as np

__author__ = 'Robbert Harms'
__date__ = '2018-08-27'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class Racket(CompartmentTemplate):
    """Computes the Racket model from [1].

    As in the original paper, we optimize "kw = k1/k2"  to preserve the inequality k1 >= k2 during optimization.

    References:
    [1] Sotiropoulos SN, Behrens TEJ, Jbabdi S. Ball and rackets: Inferring fiber fanning from
        diffusion-weighted MRI. Neuroimage. 2012;60(2):1412-1425. doi:10.1016/j.neuroimage.2012.01.056.
    """
    parameters = ('g', 'b', 'd', 'theta', 'phi', 'psi', 'k1', 'kw', '@cache')
    dependencies = ['eigenvalues_3x3_symmetric', 'ConfluentHyperGeometricFirstKind']
    cl_code = '''
        /**
         *  This computes Q = R.T*B_diag*R * -bdg^2 in one go. 
         
            This code was generated using sympy and some manual adjustments::

                from sympy import symbols, Matrix, simplify, cos, sin
                psi, phi, theta, k1, k2 = symbols("psi, phi, theta, k1, k2")
    
                R_psi = Matrix([[cos(psi), sin(psi), 0], [-sin(psi), cos(psi), 0], [0, 0, 1]])
                R_theta = Matrix([[cos(theta), 0, -sin(theta)], [0, 1, 0], [sin(theta), 0, cos(theta)]])
                R_phi = Matrix([[cos(phi), sin(phi), 0], [-sin(phi), cos(phi), 0], [0, 0, 1]])
                B_diag = Matrix([[-k1, 0, 0], [0, -k2, 0], [0, 0, 0]])
    
                R = R_psi * R_theta * R_phi
                B = simplify(R.T * B_diag * R)

            Please note that Q = B-bdg^2 is supposed to be symmetric.
        */
        double Q[6]; // upper triangular
        Q[0] = -b*d*g.x*g.x + cache->B[0];
        Q[1] = -b*d*g.x*g.y + cache->B[1];
        Q[2] = -b*d*g.x*g.z + cache->B[2];
        Q[3] = -b*d*g.y*g.y + cache->B[3];
        Q[4] = -b*d*g.y*g.z + cache->B[4];
        Q[5] = -b*d*g.z*g.z + cache->B[5];

        double e[3];
        eigenvalues_3x3_symmetric(Q, e);
        
        return ConfluentHyperGeometricFirstKind(-e[0], -e[1], -e[2]) / *cache->denom;
    '''
    cache_info = {
        'fields': ['double denom', ('double', 'B', 6)],
        'cl_code': '''
            double k2 = k1 / kw;
            
            double cos_theta;
            double sin_theta = sincos(theta, &cos_theta);
            double cos_phi;
            double sin_phi = sincos(phi, &cos_phi);
            double cos_psi;
            double sin_psi = sincos(psi, &cos_psi);
            
            cache->B[0] = -k1*pown(sin_phi*sin_psi - cos_phi*cos_psi*cos_theta, 2) 
                            - k2*pown(sin_phi*cos_psi + sin_psi*cos_phi*cos_theta, 2);
            cache->B[1] = k1*(sin_phi*sin_psi - cos_phi*cos_psi*cos_theta)*(sin_phi*cos_psi*cos_theta + sin_psi*cos_phi) 
                            - k2*(sin_phi*cos_psi + sin_psi*cos_phi*cos_theta)
                                    *(sin_phi*sin_psi*cos_theta - cos_phi*cos_psi);
            cache->B[2] = (-k1*(sin_phi*sin_psi - cos_phi*cos_psi*cos_theta)*cos_psi 
                            + k2*(sin_phi*cos_psi + sin_psi*cos_phi*cos_theta)*sin_psi)*sin_theta;
            cache->B[3] = -k1*pown(sin_phi*cos_psi*cos_theta + sin_psi*cos_phi, 2) 
                            - k2*pown(sin_phi*sin_psi*cos_theta - cos_phi*cos_psi, 2);
            cache->B[4] = (k1*(sin_phi*cos_psi*cos_theta + sin_psi*cos_phi)*cos_psi 
                            + k2*(sin_phi*sin_psi*cos_theta - cos_phi*cos_psi)*sin_psi)*sin_theta;
            cache->B[5] = -(k1*cos_psi*cos_psi + k2*sin_psi*sin_psi)*sin_theta*sin_theta;
            
            *cache->denom = ConfluentHyperGeometricFirstKind(k1, k2, 0);
        '''
    }
    extra_optimization_maps = [
        lambda d: {'k2': d['k1'] / d['kw']},
        lambda d: {'ODI_k1': np.arctan2(1.0, d['k1']) * 2 / np.pi,
                   'ODI_k2': np.arctan2(1.0, d['k2']) * 2 / np.pi,
                   'ODI': np.sqrt(np.abs(np.arctan2(1.0, (d['k1'] * d['k2'])) * 2 / np.pi))}
    ]
    extra_sampling_maps = [
        lambda samples: {'k1': np.mean(samples['k1'], axis=1),
                         'k1.std': np.std(samples['k1'], axis=1),
                         'k2': np.mean(samples['k1'] / samples['kw'], axis=1),
                         'k2.std': np.std(samples['k1'] / samples['kw'], axis=1)}
    ]



import numpy as np
from mdt import CompartmentTemplate, FreeParameterTemplate
from mdt.lib.post_processing import DTIMeasures
from mdt.model_building.parameter_functions.transformations import ScaleTransform

__author__ = 'Robbert Harms'
__date__ = "2015-06-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


def extra_dti_results(results_dict):
    output = {}

    matrix = np.array(
        [[results_dict['D_{}{}'.format(i, j) if 'D_{}{}'.format(i, j) in results_dict else 'D_{}{}'.format(j, i)]
          for j in range(3)] for i in range(3)]).transpose()

    eigen_values, eigen_vectors = np.linalg.eig(matrix)
    eigen_values = np.maximum(eigen_values, 0)

    output.update({'vec{}'.format(ind): eigen_vectors[:, ind] for ind in range(3)})
    output['d'] = eigen_values[:, 0]
    output['dperp0'] = eigen_values[:, 1]
    output['dperp1'] = eigen_values[:, 2]

    output.update(DTIMeasures.extra_optimization_maps(output))

    return output


class SymmetricNonParametricTensor(CompartmentTemplate):
    """The Tensor model in which a symmetric D matrix is optimized directly.

    This does not use the vector/diffusivity parameterization using the eigenvectors/eigenvalues.
    """
    parameters = ('g', 'b',
                  'D_00', 'D_01', 'D_02',
                          'D_11', 'D_12',
                                  'D_22')
    cl_code = '''
        double diff = g.x * g.x * D_00 +
                      g.y * g.x * D_01 * 2 +
                      g.y * g.y * D_11 +
                      g.z * g.x * D_02 * 2 +
                      g.z * g.y * D_12 * 2 +
                      g.z * g.z * D_22;
        return exp(-b * diff);
    '''
    extra_optimization_maps = [extra_dti_results]

    class D_00(FreeParameterTemplate):
        init_value = 0.3e-9
        lower_bound = 0
        upper_bound = 5e-9
        parameter_transform = ScaleTransform(1e10)
        sampling_proposal_std = 1e-10
        numdiff_info = {'scale_factor': 1e10, 'use_upper_bound': False}

    class D_11(FreeParameterTemplate):
        init_value = 0.3e-9
        lower_bound = 0
        upper_bound = 5e-9
        parameter_transform = ScaleTransform(1e10)
        sampling_proposal_std = 1e-10
        numdiff_info = {'scale_factor': 1e10, 'use_upper_bound': False}

    class D_22(FreeParameterTemplate):
        init_value = 1.2e-9
        lower_bound = 0
        upper_bound = 5e-9
        parameter_transform = ScaleTransform(1e10)
        sampling_proposal_std = 1e-10
        numdiff_info = {'scale_factor': 1e10, 'use_upper_bound': False}

    class D_01(FreeParameterTemplate):
        init_value = 0.1e-9
        lower_bound = 0
        upper_bound = 5e-9
        parameter_transform = ScaleTransform(1e10)
        sampling_proposal_std = 1e-10
        numdiff_info = {'scale_factor': 1e10, 'use_upper_bound': False}

    class D_02(FreeParameterTemplate):
        init_value = 0.1e-9
        lower_bound = 0
        upper_bound = 5e-9
        parameter_transform = ScaleTransform(1e10)
        sampling_proposal_std = 1e-10
        numdiff_info = {'scale_factor': 1e10, 'use_upper_bound': False}

    class D_12(FreeParameterTemplate):
        init_value = 0.1e-9
        lower_bound = 0
        upper_bound = 5e-9
        parameter_transform = ScaleTransform(1e10)
        sampling_proposal_std = 1e-10
        numdiff_info = {'scale_factor': 1e10, 'use_upper_bound': False}



from mdt import CompositeModelTemplate

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class Kurtosis(CompositeModelTemplate):
    """The standard Kurtosis model with in vivo defaults."""
    model_expression = '''
        S0 * KurtosisTensor
    '''
    volume_selection = {'b': [(0, 3.0e9 + 0.1e9)]}



from mot.lib.cl_function import SimpleCLFunctionParameter
from .parameter_functions.numdiff_info import SimpleNumDiffInfo
from .parameter_functions.priors import UniformWithinBoundsPrior
from .parameter_functions.transformations import IdentityTransform
import numpy as np

__author__ = 'Robbert Harms'
__date__ = "2016-10-03"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class InputDataParameter(SimpleCLFunctionParameter):

    def __init__(self, declaration, value):
        """These parameters signal are meant to be contain data loaded from the input data object.

        In contrast to free parameters which are being optimized (or fixed to values), these parameters are
        meant to be loaded from the input data. They can contain scalars, vectors or matrices with values
        to use for each problem instance and each data point.

        Args:
            declaration (str): the declaration of this parameter. For example ``global int foo``.
            value (double or ndarray): The value used if no value is given in the input data.
        """
        super().__init__(declaration)
        self.value = value


class ProtocolParameter(InputDataParameter):

    def __init__(self, declaration, value=None):
        """Caries data per observation.

        Values for this parameter type are typically loaded from the input data. A default can be provided in the case
        that there is no suitable value in the input data.

        Args:
            declaration (str): the declaration of this parameter. For example ``global int foo``.
            value (None or float or ndarray): The value used if no value is given in the input data.
        """
        super().__init__(declaration, value=value)


class FreeParameter(SimpleCLFunctionParameter):

    def __init__(self, declaration, fixed, value, lower_bound, upper_bound,
                 parameter_transform=None, sampling_proposal_std=None,
                 sampling_prior=None, numdiff_info=None):
        """This are the kind of parameters that are generally meant to be optimized.

        Args:
            declaration (str): the declaration of this parameter. For example ``global int foo``.
            fixed (boolean): If this parameter is fixed to the value given
            value (double or ndarray): A single value for all problems or a list of values for each problem.
            lower_bound (double): The lower bound of this parameter
            upper_bound (double): The upper bound of this parameter
            parameter_transform (mdt.model_building.parameter_functions.transformations.AbstractTransformation):
                The parameter transformation function
            sampling_proposal_std (float): The proposal standard deviation, used in some MCMC sample routines
            sampling_prior (mdt.model_building.parameter_functions.priors.ParameterPrior): The prior function for
                use in model sample
            numdiff_info (mdt.model_building.parameter_functions.numdiff_info.NumDiffInfo): the information
                for taking the numerical derivative with respect to this parameter.
        """
        super().__init__(declaration)
        self._value = value
        self._lower_bound = lower_bound
        self._upper_bound = upper_bound
        self._fixed = fixed

        self._parameter_transform = parameter_transform or IdentityTransform()
        self._sampling_proposal_std = sampling_proposal_std or 1
        self._sampling_prior = sampling_prior or UniformWithinBoundsPrior()
        self._numdiff_info = numdiff_info or SimpleNumDiffInfo()

    @property
    def value(self):
        return self._value

    @property
    def lower_bound(self):
        return self._lower_bound

    @property
    def upper_bound(self):
        return self._upper_bound

    @property
    def fixed(self):
        return self._fixed

    @property
    def parameter_transform(self):
        """Get the parameter transformation function used during optimization.

        Returns:
            mdt.model_building.parameter_functions.transformations.AbstractTransformation: the transformation method
        """
        return self._parameter_transform

    @property
    def sampling_proposal_std(self):
        """Get the initial proposal standard deviation for this parameter.

        Returns:
            float: the initial default proposal standard deviation for use in MCMC sampling
        """
        return self._sampling_proposal_std

    @property
    def sampling_prior(self):
        """Get the prior for this parameter, this is used in MCMC sampling.

        Returns:
            mdt.model_building.parameter_functions.priors.ParameterPrior: the prior for this parameter
        """
        return self._sampling_prior

    @property
    def numdiff_info(self):
        """Specifies how to numerically differentiate this parameter.

        Returns:
            mdt.model_building.parameter_functions.numdiff_info.NumDiffInfo: the numerical differentiation information
        """
        return self._numdiff_info


class SphericalCoordinateParameter(FreeParameter):
    """Template base class for spherical coordinate parameters.

    These are meant to be inherited by the polar angle class and the azimuth angle class.
    """


class PolarAngleParameter(SphericalCoordinateParameter):
    """Polar angle for use in spherical coordinate systems.

    If a compartment uses both a :class:`PolarAngleParameter` and :class:`AzimuthAngleParameter`,
    the composite model will ensure that the resulting cartesian coordinates are within the right spherical hemisphere.
    This is possible since diffusion is symmetric.

    In the background, we limit both the the polar angle and the azimuth angle between [0, pi] parameter
    between [0, pi] by projecting any other angle combination onto the right spherical hemisphere.
    """


class AzimuthAngleParameter(SphericalCoordinateParameter):
    """Azimuth angle for use in spherical coordinate systems.

    If a compartment uses both a :class:`PolarAngleParameter` and :class:`AzimuthAngleParameter`,
    the composite model will ensure that the resulting cartesian coordinates are within the right spherical hemisphere.
    This is possible since diffusion is symmetric.

    In the background, we limit both the the polar angle and the azimuth angle between [0, pi] parameter
    between [0, pi] by projecting any other angle combination onto the right spherical hemisphere.
    """


class RotationalAngleParameter(FreeParameter):

    def __init__(self, *args, modulus=np.pi, **kwargs):
        """Base class for parameters for which we want to enforce a modulus range.

        Parameters of this type are essentially unbounded, but their range is restricted to [0, modulus] using a modulo
        transformation. The modulus can be provided as an argument. This parameter class is recognized by the
        composite model which adds the necessary functions to the optimization and sampling routines.
        """
        super().__init__(*args, **kwargs)
        self.modulus = modulus


class NoiseStdFreeParameter(FreeParameter):
    """Specifies that this parameter should be set to the current noise standard deviation estimate.

    Parameters of this type are only meant to be used in the likelihood functions. They indicate the
    parameter to use for the initialization of the noise standard deviation.
    """


class NoiseStdInputParameter(SimpleCLFunctionParameter):

    def __init__(self, name='noise_std'):
        """Parameter indicating that this parameter should be fixed to the current value of the noise std.

        Parameters of this type are meant to be used in compartment models specifying that we should use the
        noise standard deviation value from the likelihood function as input to this function.
        """
        super().__init__('double ' + name)


class LibraryParameter(SimpleCLFunctionParameter):
    """Parameters of this type are used inside library functions. They are not meant to be used in Model functions.
    """


class CurrentObservationParam(SimpleCLFunctionParameter):

    def __init__(self, name='observation'):
        """This parameter indicates that the model should inject the current observation value in the model.

        Sometimes during model linearization or other mathematical operations the current observation appears on
        both sides of the optimization equation. That is, it sometimes happens you want to use the current observation
        to model that same observation. This parameter is a signal to the model builder to inject the current
        observation.

        You can use this parameter by adding it to your model and then use the current name in your model equation.
        """
        super().__init__('double ' + name)


class CurrentModelSignalParam(SimpleCLFunctionParameter):

    def __init__(self, name='model_signal'):
        """This parameter indicates that the model should inject here the current signal value of the model.

        Parameters of this type can only be used by the signal noise and the likelihood functions.
        """
        super().__init__('double ' + name)


class DataCacheParameter(SimpleCLFunctionParameter):

    def __init__(self, compartment_name, name):
        """This class provides a subclass for checking instance types.

        Args:
            compartment_name (str): the name of the compartment holding this parameter.
                This parameter will make sure it gets named to the correct caching struct type.
            name (str): the name of this parameter in the function
        """
        super().__init__('{}_cache* {}'.format(compartment_name, name))



from contextlib import contextmanager

__author__ = 'Robbert Harms'
__date__ = "2015-10-27"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class EstimableModel:

    def __init__(self, *args, **kwargs):
        """This is an interface for all methods needed to be able to optimize and sample a model.
        """
        super().__init__()

    @property
    def name(self):
        """The name of this model.

        Returns:
            str: the name of this model.
        """
        raise NotImplementedError()

    @contextmanager
    def voxels_to_analyze_context(self, voxels_to_analyze):
        """Context manager that should temporarily set the voxels we are analyzing.

        Args:
            voxels_to_analyze (List[int]): list of (1d ROI) voxel indices with the voxels we want to compute.
        """
        raise NotImplementedError()

    def set_input_data(self, input_data):
        """Set the input data this model will deal with.

        Args:
            input_data (mdt.utils.MRIInputData): The container for the data we will use for this model.

        Returns:
            Returns self for chainability
        """
        raise NotImplementedError()

    def get_kernel_data(self):
        """Get the kernel data this model needs for evaluation in OpenCL.

        This is needed for evaluating the priors, likelihoods and other functions.

        Returns:
            mot.lib.kernel_data.KernelData: the kernel data used by this model
        """
        raise NotImplementedError()

    def get_objective_function(self):
        """For minimization, get the negative of the log-likelihood function.

        Returns:
            mot.lib.cl_function.CLFunction: the CL function for the optimization routines in MOT.
        """
        raise NotImplementedError()

    def get_constraints_function(self):
        """The function for the (inequality) constraints.

        Returns:
            mot.lib.cl_function.CLFunction: the CL function for the inequality constraints of this model.
        """
        raise NotImplementedError()

    def get_log_likelihood_function(self):
        """For sampling, get the log-likelihood function.

        Returns:
            mot.lib.cl_function.CLFunction: the CL function for the log likelihood function during MCMC sampling
        """
        raise NotImplementedError()

    def get_log_prior_function(self):
        """Get the prior function used during sampling.

        Returns:
            mot.lib.cl_function.CLFunction: the CL function for the log prior function during MCMC sampling
        """
        raise NotImplementedError()

    def get_finalize_proposal_function(self):
        """Get the function used to finalize the proposal.

        This function is used to finalize the proposals during sampling.

        Returns:
            mot.lib.cl_function.CLFunction: the CL function used to finalize a proposal during sampling.
        """
        raise NotImplementedError()

    def get_initial_parameters(self):
        """Get the initial parameters for eac of the voxels in ``voxels_to_analyze``.

        Returns:
            ndarray: 2d array with for every problem (first dimension) the initial parameters (second dimension).
        """
        raise NotImplementedError()

    def get_post_optimization_output(self, optimized_parameters, return_codes):
        """Get the output after optimization.

        This is called by the processing strategy to finalize the optimization of a batch of voxels.

        Returns:
            dict: dictionary with results maps, can be nested which should translate to sub-directories.
        """
        raise NotImplementedError()

    def get_rwm_proposal_stds(self):
        """Get the Random Walk Metropolis proposal standard deviations for every parameter and every problem instance.

        These proposal standard deviations are used in Random Walk Metropolis MCMC sample.

        Returns:
            ndarray: the proposal standard deviations of each free parameter, for each problem instance
        """
        raise NotImplementedError()

    def get_rwm_epsilons(self):
        """Get per parameter a value small relative to the parameter's standard deviation.

        This is used in, for example, the SCAM Random Walk Metropolis sampling routine to add to the new proposal
        standard deviation to ensure it does not collapse to zero.

        Returns:
            list: per parameter an epsilon, relative to the proposal standard deviation
        """
        raise NotImplementedError()

    def get_random_parameter_positions(self, nmr_positions=1):
        """Get one or more random parameter positions centered around the initial parameters.

        This can be used to generate random starting points for sampling routines with multiple walkers.

        Per position requested, this function generates a normal distribution around the initial parameters (using
        :meth:`get_initial_parameters`) with the standard deviation derived from the random walk metropolis std.

        Returns:
            ndarray: a 3d matrix for (voxels, parameters, nmr_positions).
        """
        raise NotImplementedError()

    def post_process_optimization_maps(self, results_dict, results_array=None, log_likelihoods=None):
        """Compute additional result maps.

        This function behaves as a procedure and as a function. The input dict can be updated in place, but it should
        also return a dict but that is merely for the purpose of chaining.

        Args:
            results_dict (dict): A dictionary with as keys the names of the parameters and as values the 1d maps with
                for each voxel the optimized parameter value. The given dictionary can be altered by this function.
            results_array (ndarray): if available, the results as an array instead of as a dictionary, if not given we
                will construct it in this function.
            log_likelihoods (ndarray): for every set of parameters the corresponding log likelihoods.
                If not provided they will be calculated from the parameters.

        Returns:
            dict: The same result dictionary but with updated values or with additional maps.
                It should at least return the results_dict.
        """
        raise NotImplementedError()

    def get_post_sampling_maps(self, sampling_output):
        """Get the post sampling volume maps.

        This will return a dictionary mapping folder names to dictionaries with volumes to write.

        Args:
            sampling_output (mot.sample.base.SamplingOutput): the output of the sampler

        Returns:
            dict: a dictionary with for every subdirectory the maps to save
        """
        raise NotImplementedError()

    def get_nmr_observations(self):
        """Get the number of observations in the data.

        Returns:
            int: the number of observations present in the data
        """
        raise NotImplementedError()

    def get_nmr_parameters(self):
        """Get the number of estimable parameters in this model.

        Returns:
            int: the number of estimable parameters
        """
        raise NotImplementedError()

    def get_lower_bounds(self):
        """Get the lower bounds.

        Returns:
            List: for every parameter a lower bound which can either be None, a scalar or a vector with a
                lower bound per problem instance.
        """
        raise NotImplementedError()

    def get_upper_bounds(self):
        """Get the upper bounds.

        Returns:
            List: for every parameter an upper bound which can either be None, a scalar or a vector with a
                upper bound per problem instance.
        """
        raise NotImplementedError()

    def get_free_param_names(self):
        """Get the names of the free parameters.

        Returns:
            List: the name of the free parameters
        """
        raise NotImplementedError()

    def get_parameter_codec(self):
        """Get a parameter codec that can be used to transform the parameters to and from optimization and model space.

        Returns:
            mdt.model_building.utils.ParameterCodec: an instance of a parameter codec
        """
        raise NotImplementedError()

    def update_active_post_processing(self, processing_type, settings):
        """Update the active post-processing semaphores.

        It is possible to control which post-processing routines get run by overwriting them using this method.
        For a list of post-processors, please see the default mdt configuration file under ``active_post_processing``.

        Args:
            processing_type (str): one of ``sample`` or ``optimization``.
            settings (dict): the items to set in the post-processing information
        """
        raise NotImplementedError()

    def is_input_data_sufficient(self, input_data=None):
        """Check if the input data has enough information for this model to work.

        Args:
            input_data (mdt.utils.MRIInputData): The input data we intend on using with this model.

        Returns:
            boolean: True if there is enough information in the input data, false otherwise.
        """
        raise NotImplementedError()

    def get_input_data_problems(self, input_data=None):
        """Get all the problems with the protocol.

        Args:
            input_data (mdt.utils.MRIInputData): The input data we intend on using with this model.

        Returns:
            list of InputDataProblem: A list of
                InputDataProblem instances or subclasses of that baseclass.
                These objects indicate the problems with the protocol and this model.
        """
        raise NotImplementedError()


class InputDataProblem:

    def __init__(self):
        """The base class for indicating problems with the input data.

        These are meant to be returned from the function get_input_data_problems().

        Each of these problems is supposed to overwrite the function __str__() for reporting the problem.
        """

    def __repr__(self):
        return self.__str__()


class MissingProtocolInput(InputDataProblem):

    def __init__(self, missing_columns):
        super().__init__()
        self.missing_columns = missing_columns

    def __str__(self):
        return 'Missing columns: ' + ', '.join(self.missing_columns)


class NamedProtocolProblem(InputDataProblem):

    def __init__(self, model_protocol_problem, model_name):
        """This extends the given model protocol problem to also include the name of the model.

        Args:
            model_protocol_problem (InputDataProblem): The name for the problem with the given model.
            model_name (str): the name of the model
        """
        super().__init__()
        self._model_protocol_problem = model_protocol_problem
        self._model_name = model_name

    def __str__(self):
        return "{0}: {1}".format(self._model_name, self._model_protocol_problem)



from mdt import CompositeModelTemplate

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class S0(CompositeModelTemplate):
    """Models the unweighted signal (aka. b0)."""
    model_expression = 'S0'
    volume_selection = {'b': [(0, 250e6)]}



import logging
import numpy as np
from scipy.ndimage import binary_dilation, generate_binary_structure, binary_fill_holes
from mdt.utils import load_brain_mask
from mdt.protocols import load_protocol
from mdt.lib.nifti import load_nifti, write_nifti
import mot.configuration
from scipy.ndimage.filters import median_filter

__author__ = 'Robbert Harms'
__date__ = "2015-07-20"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


def create_median_otsu_brain_mask(dwi_info, protocol, mask_threshold=0, fill_holes=True, **kwargs):
    """Create a brain mask using the given volume.

    Args:
        dwi_info (string or tuple or image): The information about the volume, either:

            - the filename of the input file
            - or a tuple with as first index a ndarray with the DWI and as second index the header
            - or only the image as an ndarray
        protocol (string or :class:`~mdt.protocols.Protocol`): The filename of the protocol file or a Protocol object
        mask_threshold (float): everything below this b-value threshold is masked away (value in s/m^2)
        fill_holes (boolean): if we will fill holes after the median otsu algorithm and before the thresholding
        **kwargs: the additional arguments for median_otsu.

    Returns:
        ndarray: The created brain mask
    """
    logger = logging.getLogger(__name__)
    logger.info('Starting calculating a brain mask')

    if isinstance(dwi_info, str):
        signal_img = load_nifti(dwi_info)
        dwi = signal_img.get_data()
    elif isinstance(dwi_info, (tuple, list)):
        dwi = dwi_info[0]
    else:
        dwi = dwi_info

    if isinstance(protocol, str):
        protocol = load_protocol(protocol)

    if len(dwi.shape) == 4:
        unweighted_ind = protocol.get_unweighted_indices()
        if len(unweighted_ind):
            unweighted = np.mean(dwi[..., unweighted_ind], axis=3)
        else:
            unweighted = np.mean(dwi, axis=3)
    else:
        unweighted = dwi.copy()

    brain_mask = median_otsu(unweighted, **kwargs)
    brain_mask = brain_mask > 0

    if fill_holes:
        brain_mask = binary_fill_holes(brain_mask)

    if mask_threshold:
        brain_mask = np.mean(dwi[..., protocol.get_weighted_indices()], axis=3) * brain_mask > mask_threshold

    logger.info('Finished calculating a brain mask')

    return brain_mask


def generate_simple_wm_mask(scalar_map, whole_brain_mask, threshold=0.3, median_radius=1, nmr_filter_passes=2):
    """Generate a simple white matter mask by thresholding the given map and smoothing it using a median filter.

    Everything below the given threshold will be masked (not used). It also applies the regular brain mask to
    only retain values inside the brain.

    Args:
        scalar_map (str or ndarray): the path to the FA file
        whole_brain_mask (str or ndarray): the general brain mask used in the FA model fitting
        threshold (double): the FA threshold. Everything below this threshold is masked (set to 0). To be precise:
            where fa_data < fa_threshold set the value to 0.
        median_radius (int): the radius of the median filter
        nmr_filter_passes (int): the number of passes we apply the median filter
    """
    filter_footprint = np.zeros((1 + 2 * median_radius,) * 3)
    filter_footprint[median_radius, median_radius, median_radius] = 1
    filter_footprint[:, median_radius, median_radius] = 1
    filter_footprint[median_radius, :, median_radius] = 1
    filter_footprint[median_radius, median_radius, :] = 1

    if isinstance(scalar_map, str):
        map_data = load_nifti(scalar_map).get_data()
    else:
        map_data = np.copy(scalar_map)

    map_data[map_data < threshold] = 0
    wm_mask = map_data.astype(np.bool)

    if len(wm_mask.shape) > 3:
        wm_mask = wm_mask[:, :, :, 0]

    wm_mask[np.logical_not(load_brain_mask(whole_brain_mask))] = 0

    if nmr_filter_passes == 0:
        return wm_mask

    mask = load_brain_mask(whole_brain_mask)

    wm_mask_masked = np.ma.masked_array(wm_mask, mask=mask)
    for ind in range(nmr_filter_passes):
        wm_mask_masked = median_filter(wm_mask_masked, footprint=filter_footprint, mode='constant')
    return wm_mask_masked


def create_write_median_otsu_brain_mask(dwi_info, protocol, output_fname, **kwargs):
    """Write a brain mask using the given volume and output as the given volume.

    Args:
        dwi_info (string or tuple or ndarray): the filename of the input file or a tuple with as
            first index a ndarray with the DWI and as second index the header or only the image.
        protocol (string or :class:`~mdt.protocols.Protocol`): The filename of the protocol file or a Protocol object
        output_fname (string): the filename of the output file (the extracted brain mask)
            If None, no output is written. If ``dwi_info`` is an ndarray also no file is written
            (we don't have the header).

    Returns:
        ndarray: The created brain mask
    """
    if isinstance(dwi_info, str):
        signal_img = load_nifti(dwi_info)
        dwi = signal_img.get_data()
        header = signal_img.header
    else:
        dwi = dwi_info[0]
        header = dwi_info[1]

    mask = create_median_otsu_brain_mask(dwi, protocol, **kwargs)
    write_nifti(mask, output_fname, header)

    return mask


def median_otsu(unweighted_volume, median_radius=4, numpass=4, dilate=1):
    """ Simple brain extraction tool for dMRI data.

    This function is inspired from the ``median_otsu`` function from ``dipy``
    and is copied here to remove a dependency.

    It uses a median filter smoothing of the ``unweighted_volume``
    automatic histogram Otsu thresholding technique, hence the name
    *median_otsu*.

    This function is inspired from Mrtrix's bet which has default values
    ``median_radius=3``, ``numpass=2``. However, from tests on multiple 1.5T
    and 3T data. From GE, Philips, Siemens, the most robust choice is
    ``median_radius=4``, ``numpass=4``.

    Args:
        unweighted_volume (ndarray): ndarray of the unweighted volumes brain volumes
        median_radius (int): Radius (in voxels) of the applied median filter (default 4)
        numpass (int): Number of pass of the median filter (default 4)
        dilate (None or int): optional number of iterations for binary dilation

    Returns:
        ndarray: a 3D ndarray with the binary brain mask
    """
    b0vol = unweighted_volume

    logger = logging.getLogger(__name__)
    logger.info('We will use a single precision float type for the calculations.'.format())
    for env in mot.configuration.get_cl_environments():
        logger.info('Using device \'{}\'.'.format(str(env)))

    for ind in range(numpass):
        b0vol = median_filter(b0vol, size=median_radius, mode='mirror')

    thresh = _otsu(b0vol)
    mask = b0vol > thresh

    if dilate is not None:
        cross = generate_binary_structure(3, 1)
        mask = binary_dilation(mask, cross, iterations=dilate)

    return mask


def _otsu(image, nbins=256):
    """
    Return threshold value based on Otsu's method.
    Copied from scikit-image to remove dependency.

    Parameters
    ----------
    image : array
        Input image.
    nbins : int
        Number of bins used to calculate histogram. This value is ignored for
        integer arrays.

    Returns
    -------
    threshold : float
        Threshold value.
    """
    hist, bin_centers = np.histogram(image, nbins)
    hist = hist.astype(np.float)

    # class probabilities for all possible thresholds
    weight1 = np.cumsum(hist)
    weight2 = np.cumsum(hist[::-1])[::-1]

    # class means for all possible thresholds
    mean1 = np.cumsum(hist * bin_centers[1:]) / weight1
    mean2 = (np.cumsum((hist * bin_centers[1:])[::-1]) / weight2[::-1])[::-1]

    # Clip ends to align class 1 and class 2 variables:
    # The last value of `weight1`/`mean1` should pair with zero values in
    # `weight2`/`mean2`, which do not exist.
    variance12 = weight1[:-1] * weight2[1:] * (mean1[:-1] - mean2[1:])**2

    idx = np.argmax(variance12)
    threshold = bin_centers[:-1][idx]
    return threshold



import numpy as np
from mdt.component_templates.base import ComponentBuilder, ComponentTemplate
from mdt.lib.components import has_component, get_component
from mdt.model_building.parameter_functions.numdiff_info import NumDiffInfo, SimpleNumDiffInfo
from mdt.model_building.parameters import ProtocolParameter, FreeParameter, SphericalCoordinateParameter, \
    PolarAngleParameter, AzimuthAngleParameter, RotationalAngleParameter
from mdt.model_building.parameter_functions.priors import UniformWithinBoundsPrior
from mdt.model_building.parameter_functions.transformations import AbstractTransformation


__author__ = 'Robbert Harms'
__date__ = "2015-12-12"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ParameterBuilder(ComponentBuilder):

    def _create_class(self, template):
        """Creates parameter classes

        Args:
            template (Type[ParameterTemplate]): the configuration for the parameter.
        """
        if issubclass(template, ProtocolParameterTemplate):
            class AutoProtocolParameter(ProtocolParameter):
                def __init__(self, nickname=None):
                    super().__init__('{} {}'.format(template.data_type, nickname or template.name),
                                     value=template.value)

            for name, method in template.bound_methods.items():
                setattr(AutoProtocolParameter, name, method)

            return AutoProtocolParameter

        elif issubclass(template, FreeParameterTemplate):
            numdiff_info = template.numdiff_info
            if not isinstance(numdiff_info, NumDiffInfo) and numdiff_info is not None:
                numdiff_info = SimpleNumDiffInfo(**numdiff_info)

            kwargs = dict(
                parameter_transform=_resolve_parameter_transform(template.parameter_transform),
                sampling_proposal_std=template.sampling_proposal_std,
                sampling_prior=template.sampling_prior,
                numdiff_info=numdiff_info
            )

            base_class = FreeParameter
            if issubclass(template, SphericalCoordinateParameterTemplate):
                base_class = SphericalCoordinateParameter
                if issubclass(template, PolarAngleParameterTemplate):
                    base_class = PolarAngleParameter
                if issubclass(template, AzimuthAngleParameterTemplate):
                    base_class = AzimuthAngleParameter
            elif issubclass(template, RotationalAngleParameterTemplate):
                base_class = RotationalAngleParameter
                kwargs.update(modulus=template.modulus)

            class AutoFreeParameter(base_class):
                def __init__(self, nickname=None):
                    super().__init__(
                        '{} {}'.format(template.data_type, nickname or template.name),
                        template.fixed,
                        template.init_value,
                        template.lower_bound,
                        template.upper_bound,
                        **kwargs
                    )

            for name, method in template.bound_methods.items():
                setattr(AutoFreeParameter, name, method)

            return AutoFreeParameter


class ParameterTemplate(ComponentTemplate):
    """The parameter template to inherit from.

    These templates are loaded on the fly by the ParametersBuilder

    template options:
        name (str): the name of the parameter, defaults to the class name
        description (str): the description of this parameter
        data_type (str): the data type for this parameter
    """
    _component_type = 'parameters'
    _builder = ParameterBuilder()

    name = ''
    description = ''
    data_type = 'mot_float_type'


class ProtocolParameterTemplate(ParameterTemplate):
    """The default template options for protocol parameters.

    To save memory, protocol data is loaded as a float by default.
    """
    data_type = 'float'
    value = None


class FreeParameterTemplate(ParameterTemplate):
    """The default template options for free parameters.

    Attributes:
        init_value (float): the initial value
        fixed (boolean or ndarray of float): if this parameter is fixed or not. If not fixed this should
            hold a reference to a value or a matrix
        lower_bound (float): the lower bounds, used in the parameter transform and prior
        upper_bound (float): the upper bounds, used in the parameter transform and prior
        parameter_transform
            (str or :class:`~mdt.model_building.parameter_functions.transformations.AbstractTransformation`): the
            parameter transformation, this is used for automatic range transformation of the parameters during
            optimization. See Harms 2017 NeuroImage for details. Typical elements are:

            * ``Identity``: no transformation
            * ``Positivity``: ensures the parameters are positive
            * ``Clamp``: limits the parameter between its lower and upper bounds
            * ``CosSqrClamp``: changes the range of the optimized parameters to [0, 1] and ensures boundary constraints
            * ``SinSqrClamp``: same as ``CosSqrClamp``
            * ``SqrClamp``: same as clamp but with an additional square root to change the magnitude of the range
            * ``AbsModPi``: ensures absolute modulus of the input parameters between zero and pi.
            * ``AbsModTwoPi``: ensures absolute modulus of the input parameters between zero and two pi.

        sampling_proposal_std (float): the default proposal standard deviation for this parameter. This is used
            in some MCMC sample routines.
        sampling_prior: the prior function
        numdiff_info (dict or :class:`~mdt.model_building.parameter_functions.numdiff_info.NumDiffInfo`):
            the information necessary to take the numerical derivative of a model with respect to this parameter.
            Either a dictionary with the keyword arguments to
            :class:`~mdt.model_building.parameter_functions.numdiff_info.SimpleNumDiffInfo` or an information
            object directly. If None, we use an empty dictionary. Please note that if you override this, you will have
            to specify all of the items (no automatic inheritance of sub-items).
    """
    data_type = 'mot_float_type'
    fixed = False
    init_value = 1
    lower_bound = -1e20
    upper_bound = 1e20
    parameter_transform = 'Identity'
    sampling_proposal_std = 1
    sampling_prior = UniformWithinBoundsPrior()
    numdiff_info = {'max_step': 0.1, 'scale_factor': 1, 'use_bounds': True,
                    'use_upper_bound': True, 'use_lower_bound': True}


class SphericalCoordinateParameterTemplate(FreeParameterTemplate):
    """Template base class for spherical coordinate parameters.

    These are meant to be inherited by the polar angle template and the azimuth angle template.
    """
    init_value = np.pi / 2.0
    sampling_proposal_std = 0.1
    numdiff_info = {'max_step': 0.1, 'scale_factor': 10}


class PolarAngleParameterTemplate(SphericalCoordinateParameterTemplate):
    """Polar angle for use in spherical coordinate systems.

    If a compartment uses both a :class:`PolarAngleParameterTemplate` and :class:`AzimuthAngleParameterTemplate`,
    the composite model will ensure that the resulting cartesian coordinates are within the right spherical hemisphere.
    This is possible since diffusion is symmetric.

    In the background, we limit both the the polar angle and the azimuth angle between [0, pi] parameter
    between [0, pi] by projecting any other angle combination onto the right spherical hemisphere.
    """


class AzimuthAngleParameterTemplate(SphericalCoordinateParameterTemplate):
    """Azimuth angle for use in spherical coordinate systems.

    If a compartment uses both a :class:`PolarAngleParameterTemplate` and :class:`AzimuthAngleParameterTemplate`,
    the composite model will ensure that the resulting cartesian coordinates are within the right spherical hemisphere.
    This is possible since diffusion is symmetric.

    In the background, we limit both the the polar angle and the azimuth angle between [0, pi] parameter
    between [0, pi] by projecting any other angle combination onto the right spherical hemisphere.
    """


class RotationalAngleParameterTemplate(FreeParameterTemplate):
    """Template base class for parameters for which we want to enforce a modulus range.

    Parameters of this type are essentially unbounded, but their range is restricted to [0, modulus] using a modulo
    transformation. The modulus can be provided as an argument. This parameter class is recognized by the
    composite model which adds the necessary functions to the optimization and sampling routines.
    """
    init_value = np.pi / 2.0
    modulus = np.pi
    sampling_proposal_std = 0.1
    numdiff_info = {'max_step': 0.1, 'scale_factor': 10}


def _resolve_parameter_transform(parameter_transform):
    """Resolves input parameter transforms to actual objects.

    Args:
        parameter_transform
            (str or :class:`~mdt.model_building.parameter_functions.transformations.AbstractTransformation`):
            a parameter transformation name (with or without the postfix ``Transform``) or an actual object we
            just return.

    Returns:
        mdt.model_building.parameter_functions.transformations.AbstractTransformation: an actual transformation object

    Raises:
        ValueError: if the parameter transformation could not be resolved.
    """
    if isinstance(parameter_transform, AbstractTransformation):
        return parameter_transform

    if has_component('parameter_transforms', parameter_transform):
        return get_component('parameter_transforms', parameter_transform)()

    if has_component('parameter_transforms', parameter_transform + 'Transform'):
        return get_component('parameter_transforms', parameter_transform + 'Transform')()

    raise ValueError('Could not resolve the parameter transformation "{}"'.format(parameter_transform))



from mot.lib.cl_function import SimpleCLFunction
from mot.lib.kernel_data import Array, Struct, LocalMemory
from mot.optimize.base import SimpleConstraintFunction

__author__ = 'Robbert Harms'
__date__ = '2017-05-29'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class ParameterDecodingWrapper:

    def __init__(self, nmr_parameters, decode_function):
        """Wrap the objective function and constraint function with parameter decoding.

        Args:
            nmr_parameters (int): the number of parameters to decode
            decode_function (mot.lib.cl_function.CLFunction): An OpenCL function that is used in the CL kernel to
                    transform the parameters from encoded space to model space so they can be used as input to the model.
                    The signature of the CL function is:

                    .. code-block:: c

                        void <fname>(void* data, local mot_float_type* x);
        """
        self._nmr_parameters = nmr_parameters
        self._decode_function = decode_function

    def wrap_input_data(self, input_data):
        """Wrap the input data with extra information this wrapper might need.

        Args:
            input_data (mot.lib.kernel_data.KernelData): the kernel data we will wrap

        Returns:
            mot.lib.kernel_data.KernelData: the wrapped kernel data
        """
        return Struct({'data': input_data, 'x_tmp': LocalMemory('mot_float_type', nmr_items=self._nmr_parameters)},
                      'objective_function_wrapper_data')

    def wrap_objective_function(self, objective_function):
        """Decorates the given objective function with parameter decoding.

        Args:
            objective_function (mot.lib.cl_function.CLFunction): A CL function with the signature:

                .. code-block:: c

                    double <func_name>(local mot_float_type* x,
                                       void* data,
                                       local mot_float_type* objective_list);

        Returns:
            mot.lib.cl_function.CLFunction: the wrapped objective function.
        """
        return SimpleCLFunction.from_string('''
            double wrapped_''' + objective_function.get_cl_function_name() + '''(
                    local mot_float_type* x,
                    void* data, 
                    local mot_float_type* objective_list){
                
                local mot_float_type* x_tmp = ((objective_function_wrapper_data*)data)->x_tmp;
                
                if(get_local_id(0) == 0){
                    for(uint i = 0; i < ''' + str(self._nmr_parameters) + '''; i++){
                        x_tmp[i] = x[i];
                    }
                    ''' + self._decode_function.get_cl_function_name() + '''(
                        ((objective_function_wrapper_data*)data)->data, 
                        x_tmp);
                }
                barrier(CLK_LOCAL_MEM_FENCE);
    
                return ''' + objective_function.get_cl_function_name() + '''(
                    x_tmp, ((objective_function_wrapper_data*)data)->data, objective_list);    
            }
        ''', dependencies=[objective_function, self._decode_function])

    def wrap_constraints_function(self, constraints_func):
        """Decorates the given constraint function with parameter decoding.

        Args:
            constraints_func (mot.optimize.base.ConstraintFunction): A CL function with the signature:

                .. code-block:: c

                    void <func_name>(local mot_float_type* x,
                                     void* data,
                                     local mot_float_type* constraints);

        Returns:
            mot.optimize.base.ConstraintFunction: the wrapped constraint function.
        """
        return SimpleConstraintFunction.from_string('''
            void wrapped_''' + constraints_func.get_cl_function_name() + '''(
                    local mot_float_type* x,
                    void* data, 
                    local mot_float_type* constraints){

                local mot_float_type* x_tmp = ((objective_function_wrapper_data*)data)->x_tmp;

                if(get_local_id(0) == 0){
                    for(uint i = 0; i < ''' + str(self._nmr_parameters) + '''; i++){
                        x_tmp[i] = x[i];
                    }
                    ''' + self._decode_function.get_cl_function_name() + '''(
                        ((objective_function_wrapper_data*)data)->data, 
                        x_tmp);
                }
                barrier(CLK_LOCAL_MEM_FENCE);

                ''' + constraints_func.get_cl_function_name() + '''(
                    x_tmp, ((objective_function_wrapper_data*)data)->data, constraints);    
            }
        ''', dependencies=[constraints_func, self._decode_function],
             nmr_constraints=constraints_func.get_nmr_constraints())


class ParameterCodec:

    def __init__(self, encode_func, decode_func, encode_bounds_func=None):
        """Create a parameter codec container.

        Args:
            encode_func (mot.lib.cl_function.CLFunction): An OpenCL function that is used in the CL kernel to
                transform the parameters from model space to encoded space so they can be used as input to an
                CL routine. The signature of the CL function is:

                .. code-block:: c

                    void <fname>(void* data, local mot_float_type* x);

            decode_func (mot.lib.cl_function.CLFunction): An OpenCL function that is used in the CL kernel to
                transform the parameters from encoded space to model space so they can be used as input to the model.
                The signature of the CL function is:

                .. code-block:: c

                    void <fname>(void* data, local mot_float_type* x);
            encode_bounds_func (Callable[[array, array], Tuple[array, array]]): encode the lower and upper bounds
                to bounds of the encoded parameter space. If not set, we won't encode the bounds
        """
        self._encode_func = encode_func
        self._decode_func = decode_func
        self._encode_bounds_func = encode_bounds_func

    def get_encode_function(self):
        """Get a CL function that can transform the model parameters from model space to an encoded space.

        Returns:
            mot.lib.cl_function.CLFunction: An OpenCL function that is used in the CL kernel to transform the parameters
                from model space to encoded space so they can be used as input to an CL routine.
                The signature of the CL function is:

                .. code-block:: c

                    void <fname>(void* data, local mot_float_type* x);
        """
        return self._encode_func

    def get_decode_function(self):
        """Get a CL function that can transform the model parameters from encoded space to model space.

        Returns:
            mot.lib.cl_function.CLFunction: An OpenCL function that is used in the CL kernel to transform the parameters
                from encoded space to model space so they can be used as input to the model.
                The signature of the CL function is:

                .. code-block:: c

                    void <fname>(void* data, local mot_float_type* x);
        """
        return self._decode_func

    def encode_bounds(self, lower_bounds, upper_bounds):
        """Encode the given bounds to the encoded parameter space.

        Args:
            lower_bounds (list): for each parameter the lower bound(s). Each element can either be a scalar or a vector.
            upper_bounds (list): for each parameter the upper bound(s). Each element can either be a scalar or a vector.

        Returns:
            tuple: the lower and the upper bounds, in a similar structure as the input
        """
        if self._encode_bounds_func is not None:
            return self._encode_bounds_func(lower_bounds, upper_bounds)
        return lower_bounds, upper_bounds

    def decode(self, parameters, kernel_data=None, cl_runtime_info=None):
        """Decode the given parameters using the given model.

        This transforms the data from optimization space to model space.

        Args:
            parameters (ndarray): The parameters to transform
            kernel_data (dict[str: mot.lib.utils.KernelData]): the additional data to load
            cl_runtime_info (mot.configuration.CLRuntimeInfo): the runtime information

        Returns:
            ndarray: The array with the transformed parameters.
        """
        return self._transform_parameters(self.get_decode_function(),
                                          parameters, kernel_data, cl_runtime_info=cl_runtime_info)

    def encode(self, parameters, kernel_data=None, cl_runtime_info=None):
        """Encode the given parameters using the given model.

        This transforms the data from model space to optimization space.

        Args:
            parameters (ndarray): The parameters to transform
            kernel_data (dict[str: mot.lib.utils.KernelData]): the additional data to load
            cl_runtime_info (mot.configuration.CLRuntimeInfo): the runtime information

        Returns:
            ndarray: The array with the transformed parameters.
        """
        return self._transform_parameters(self.get_encode_function(),
                                          parameters, kernel_data, cl_runtime_info=cl_runtime_info)

    def encode_decode(self, parameters, kernel_data=None, cl_runtime_info=None):
        """First apply an encoding operation and then apply a decoding operation again.

        This can be used to enforce boundary conditions in the parameters.

        Args:
            parameters (ndarray): The parameters to transform
            kernel_data (dict[str: mot.lib.utils.KernelData]): the additional data to load
            cl_runtime_info (mot.configuration.CLRuntimeInfo): the runtime information

        Returns:
            ndarray: The array with the transformed parameters.
        """
        encode_func = self.get_encode_function()
        decode_func = self.get_decode_function()

        func = SimpleCLFunction.from_string('''
            void encode_decode_parameters(void* data, local mot_float_type* x){
                ''' + encode_func.get_cl_function_name() + '''(data, x);
                ''' + decode_func.get_cl_function_name() + '''(data, x);
            }
        ''', dependencies=[encode_func, decode_func])
        return self._transform_parameters(func, parameters, kernel_data, cl_runtime_info=cl_runtime_info)

    @staticmethod
    def _transform_parameters(cl_func, parameters, kernel_data, cl_runtime_info=None):
        cl_named_func = SimpleCLFunction.from_string('''
            void transformParameterSpace(void* data, local mot_float_type* x){
                ''' + cl_func.get_cl_function_name() + '''(data, x);
            }
        ''', dependencies=[cl_func])

        kernel_data = {'data': kernel_data,
                       'x': Array(parameters, ctype='mot_float_type', mode='rw')}

        cl_named_func.evaluate(kernel_data, parameters.shape[0], cl_runtime_info=cl_runtime_info)
        return kernel_data['x'].get_data()



from mdt import LibraryFunctionTemplate

__author__ = 'Robbert Harms'
__date__ = '2018-05-02'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class NeumanCylinder(LibraryFunctionTemplate):
    """This function returns the displacement in the restricted signal attenuation, for radius R
        according to the C.H. Neuman model, equation (20) in [1].

    In spin-echo experiments, tau is typically TE/2.

    References:
    [1] Neuman CH. Spin echo of spins diffusing in a bounded medium.
        J Chem Phys. 1974;60(11):4508-4511. doi:10.1063/1.1680931.
    """
    return_type = 'double'
    parameters = ['double G', 'double tau', 'double d', 'double R']
    dependencies = ['MRIConstants', 'BesselRoots']
    cl_code = '''
        if(R == 0.0 || R < MOT_EPSILON){
            return 0;
        }

        double sum = 0;
        float alpha;
        
        #pragma unroll
        for(uint i = 0; i < bessel_roots_jnp_length; i++){
            alpha = bessel_roots_jnp[i] / R;

            sum += (pown(alpha, -4) / (bessel_roots_jnp[i] * bessel_roots_jnp[i] - 1)) * 
                    (2 * tau - 
                        (3 - 4 * exp(-alpha * alpha * d * tau) + exp(- alpha * alpha * d * 2 * tau)) 
                         / (alpha * alpha * d)
                     );
        }        

        return -((2 * GAMMA_H_SQ * (G*G)) / d) * sum;
    '''


class NeumanCylinderLongApprox(LibraryFunctionTemplate):
    """This function returns the displacement in the restricted signal attenuation, for radius R according to the
       long diffusion time approximation of the C.H. Neuman model, equation (28) in [1].

    In typical spin-echo experiments, ``tau = TE/2``.

    This approximation is valid when ``tau >> R^2 / d``.

    References:
    [1] Neuman CH. Spin echo of spins diffusing in a bounded medium.
        J Chem Phys. 1974;60(11):4508-4511. doi:10.1063/1.1680931.
    """
    return_type = 'double'
    parameters = ['double G', 'double tau', 'double d', 'double R']
    dependencies = ['MRIConstants']
    cl_code = '''
        return -( (pown(R, 4) * GAMMA_H_HZ_SQ * (G * G)) / d ) 
                    * (7 / 296.0) 
                    * (2 * tau - (99/112.0) * (pown(R, 2) / d)); 
    '''


class NeumanSphere(LibraryFunctionTemplate):
    """This function returns the displacement in spherical signal attenuation, for radius R
        according to the C.H. Neuman model, equation (18) in [1].

    In spin-echo experiments, tau is typically TE/2.

    References:
    [1] Neuman CH. Spin echo of spins diffusing in a bounded medium.
        J Chem Phys. 1974;60(11):4508-4511. doi:10.1063/1.1680931.
    """
    return_type = 'double'
    parameters = ['double G', 'double tau', 'double d', 'double R']
    dependencies = ['MRIConstants', 'BesselRoots']
    cl_code = '''
        if(R == 0.0 || R < MOT_EPSILON){
            return 0;
        }

        double sum = 0;
        float alpha;

        #pragma unroll
        for(uint i = 0; i < bessel_roots_j3_2_length; i++){
            alpha = bessel_roots_j3_2[i] / R;

            sum += (pown(alpha, -4) / (bessel_roots_j3_2[i] * bessel_roots_j3_2[i] - 2)) * 
                    (2 * tau - 
                        (3 - 4 * exp(-alpha * alpha * d * tau) + exp(- alpha * alpha * d * 2 * tau)) 
                         / (alpha * alpha * d)
                     );
        }        

        return -((2 * GAMMA_H_SQ * (G*G)) / d) * sum;
    '''


class NeumanSphereLongApprox(LibraryFunctionTemplate):
    """This function returns the displacement in the spherical signal attenuation, for radius R according to the
       long diffusion time approximation of the C.H. Neuman model, equation (27) in [1].

    In typical spin-echo experiments, ``tau = TE/2``.

    This approximation is valid when ``tau >> R^2 / d``.

    References:
    [1] Neuman CH. Spin echo of spins diffusing in a bounded medium.
        J Chem Phys. 1974;60(11):4508-4511. doi:10.1063/1.1680931.
    """
    return_type = 'double'
    parameters = ['double G', 'double tau', 'double d', 'double R']
    dependencies = ['MRIConstants']
    cl_code = '''
        return -( (pown(R, 4) * GAMMA_H_HZ_SQ * (G * G)) / d ) 
                    * (8 / 175.0) 
                    * (2 * tau - (581/840.0) * (pown(R, 2) / d)); 
    '''


class VanGelderenCylinder(LibraryFunctionTemplate):
    """This function returns the displacement in the restricted signal attenuation for radius R
       according to the Van Gelderen model [1, 2].

    References:
    [1] Gelderen V, D D, PC van Z, CT M. Evaluation of Restricted Diffusion in Cylinders.
        Phosphocreatine in Rabbit Leg Muscle. 1994. doi:10.1006/jmrb.1994.1038.
    [2] Wang LZ, Caprihan A, Fukushima E. The narrow-pulse criterion for pulsed-gradient spin-echo
        diffusion measurements. JMagnResonSerA. 1995;117(2):209-219.
    """
    return_type = 'double'
    parameters = ['double G', 'double Delta', 'double delta', 'double d', 'double R']
    dependencies = ['MRIConstants', 'BesselRoots']
    cl_code = '''
        if(R == 0.0 || R < MOT_EPSILON){
            return 0;
        }
        
        double sum = 0;
        float alpha;
        float alpha2_d;
        
        #pragma unroll
        for(uint i = 0; i < bessel_roots_jnp_length; i++){
            alpha = bessel_roots_jnp[i] / R;
            alpha2_d = d * alpha * alpha;

            sum += (2 * alpha2_d * delta
                    -  2
                    + (2 * exp(-alpha2_d * delta))
                    + (2 * exp(-alpha2_d * Delta))
                    - exp(-alpha2_d * (Delta - delta))
                    - exp(-alpha2_d * (Delta + delta)))
                        / ((alpha2_d * alpha * alpha2_d * alpha) * (bessel_roots_jnp[i] * bessel_roots_jnp[i] - 1));
        }
        return -2 * GAMMA_H_SQ * (G*G) * sum;
    '''


class VanGelderenSphere(LibraryFunctionTemplate):
    """This function returns the displacement in the spherical signal attenuation for radius R
       according to the Van Gelderen model [1].

    References:
    [1] Wang LZ, Caprihan A, Fukushima E. The narrow-pulse criterion for pulsed-gradient spin-echo
        diffusion measurements. JMagnResonSerA. 1995;117(2):209-219.
    """
    return_type = 'double'
    parameters = ['double G', 'double Delta', 'double delta', 'double d', 'double R']
    dependencies = ['MRIConstants']
    cl_code = '''
        if(R == 0.0 || R < MOT_EPSILON){
            return 0;
        }
        
        double sum = 0;
        float alpha;
        float alpha2_d;
        
        #pragma unroll
        for(uint i = 0; i < bessel_roots_j3_2_length; i++){
            alpha = bessel_roots_j3_2[i] / R;
            alpha2_d = d * alpha * alpha;

            sum += (2 * alpha2_d * delta
                    -  2
                    + (2 * exp(-alpha2_d * delta))
                    + (2 * exp(-alpha2_d * Delta))
                    - exp(-alpha2_d * (Delta - delta))
                    - exp(-alpha2_d * (Delta + delta)))
                        / ((alpha2_d * alpha * alpha2_d * alpha) * (bessel_roots_j3_2[i] * bessel_roots_j3_2[i] - 2));
        }
        return -2 * GAMMA_H_SQ * (G*G) * sum;
    '''



import numpy as np
import glob
import logging
import os
import shutil
import time
import timeit
from contextlib import contextmanager
from mdt.__version__ import __version__
from mdt.lib.nifti import get_all_nifti_data
from mdt.lib.components import get_model
from mdt.configuration import get_processing_strategy
from mdt.utils import create_roi, get_cl_devices, model_output_exists, \
    per_model_logging_context
from mdt.lib.processing_strategies import FittingProcessor, get_full_tmp_results_path
from mdt.lib.exceptions import InsufficientProtocolError
import mot.configuration
from mot.configuration import config_context as mot_config_context

__author__ = 'Robbert Harms'
__date__ = "2015-05-01"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


def get_optimization_inits(model_name, input_data, output_folder, cl_device_ind=None):
    """Get better optimization starting points for the given model.

    Since initialization can make quite a difference in optimization results, this function can generate
    a good initialization starting point for the given model. The idea is that before you call the :func:`fit_model`
    function, you call this function to get a better starting point. An usage example would be::

        input_data = mdt.load_input_data(..)

        init_data = get_optimization_inits('BallStick_r1', input_data, '/my/folder')

        fit_model('BallStick_r1', input_data, '/my/folder',
                  initialization_data={'inits': init_data})

    Where the init data returned by this function can directly be used as input to the ``initialization_data``
    argument of the :func`fit_model` function.

    Please note that his function only supports models shipped by default with MDT.

    Args:
        model_name (str):
            The name of a model for which we want the optimization starting points.
        input_data (:class:`~mdt.utils.MRIInputData`): the input data object containing all
            the info needed for model fitting of intermediate models.
        output_folder (string): The path to the folder where to place the output, we will make a subdir with the
            model name in it.
        cl_device_ind (int or list): the index of the CL device to use. The index is from the list from the function
            utils.get_cl_devices(). This can also be a list of device indices.

    Returns:
        dict: a dictionary with initialization points for the selected model
    """
    logger = logging.getLogger(__name__)

    def get_subset(param_names, fit_results):
        return {key: value for key, value in fit_results.items() if key in param_names}

    def get_model_fit(model_name):
        logger.info('Starting intermediate optimization for generating initialization point.')

        from mdt import fit_model
        results = fit_model(model_name, input_data, output_folder, recalculate=False, use_cascaded_inits=False,
                            initialization_data={'inits': get_init_data(model_name)})

        logger.info('Finished intermediate optimization for generating initialization point.')
        return results

    def get_init_data(model_name):
        inits = {}
        free_parameters = get_model(model_name)().get_free_param_names()

        if 'S0.s0' in free_parameters and input_data.has_input_data('b'):
            unweighted_locations = np.where(input_data.get_input_data('b') < 250e6)[0]
            inits['S0.s0'] = np.mean(input_data.signal4d[..., unweighted_locations], axis=-1)

        if model_name.startswith('BallStick_r2'):
            inits.update(get_subset(free_parameters, get_model_fit('BallStick_r1')))
            inits['w_stick1.w'] = np.minimum(inits['w_stick0.w'], 0.05)
        elif model_name.startswith('BallStick_r3'):
            inits.update(get_subset(free_parameters, get_model_fit('BallStick_r2')))
            inits['w_stick2.w'] = np.minimum(inits['w_stick1.w'], 0.05)
        elif model_name.startswith('Tensor'):
            fit_results = get_model_fit('BallStick_r1')
            inits.update(get_subset(free_parameters, fit_results))
            inits['Tensor.theta'] = fit_results['Stick0.theta']
            inits['Tensor.phi'] = fit_results['Stick0.phi']
        elif model_name.startswith('NODDI'):
            fit_results = get_model_fit('BallStick_r1')
            inits.update(get_subset(free_parameters, fit_results))
            inits['w_ic.w'] = fit_results['w_stick0.w'] / 2.0
            inits['w_ec.w'] = fit_results['w_stick0.w'] / 2.0
            inits['w_csf.w'] = fit_results['w_ball.w']
            inits['NODDI_IC.theta'] = fit_results['Stick0.theta']
            inits['NODDI_IC.phi'] = fit_results['Stick0.phi']
        elif model_name.startswith('BinghamNODDI_r1'):
            noddi_results = get_model_fit('NODDI')
            inits.update(get_subset(free_parameters, noddi_results))
            inits['w_in0.w'] = noddi_results['w_ic.w']
            inits['w_en0.w'] = noddi_results['w_ec.w']
            inits['w_csf.w'] = noddi_results['w_csf.w']
            inits['BinghamNODDI_IN0.theta'] = noddi_results['NODDI_IC.theta']
            inits['BinghamNODDI_IN0.phi'] = noddi_results['NODDI_IC.phi']
            inits['BinghamNODDI_IN0.k1'] = noddi_results['NODDI_IC.kappa']
        elif model_name.startswith('BinghamNODDI_r2'):
            bs2_results = get_model_fit('BallStick_r2')
            inits.update(get_subset(free_parameters, bs2_results))
            inits.update(get_subset(free_parameters, get_model_fit('BinghamNODDI_r1')))
            inits['BinghamNODDI_IN1.theta'] = bs2_results['Stick1.theta']
            inits['BinghamNODDI_IN1.phi'] = bs2_results['Stick1.phi']
        elif model_name.startswith('Kurtosis'):
            fit_results = get_model_fit('Tensor')
            inits.update(get_subset(free_parameters, fit_results))
            inits.update({'KurtosisTensor.' + key: fit_results['Tensor.' + key]
                          for key in ['theta', 'phi', 'psi', 'd', 'dperp0', 'dperp1']})
        elif model_name.startswith('CHARMED_r'):
            nmr_dir = model_name[len('CHARMED_r'):len('CHARMED_r')+1]
            fit_results = get_model_fit('BallStick_r' + nmr_dir)
            inits.update(get_subset(free_parameters, fit_results))
            inits['Tensor.theta'] = fit_results['Stick0.theta']
            inits['Tensor.phi'] = fit_results['Stick0.phi']
            for dir_ind in range(int(nmr_dir)):
                inits['w_res{}.w'.format(dir_ind)] = fit_results['w_stick{}.w'.format(dir_ind)]
                inits['CHARMEDRestricted{}.theta'.format(dir_ind)] = fit_results['Stick{}.theta'.format(dir_ind)]
                inits['CHARMEDRestricted{}.phi'.format(dir_ind)] = fit_results['Stick{}.phi'.format(dir_ind)]
        elif model_name.startswith('BallRacket_r'):
            nmr_dir = model_name[len('BallRacket_r'):len('BallRacket_r')+1]
            fit_results = get_model_fit('BallStick_r' + nmr_dir)
            inits.update(get_subset(free_parameters, fit_results))
            for dir_ind in range(int(nmr_dir)):
                inits['w_res{}.w'.format(dir_ind)] = fit_results['w_stick{}.w'.format(dir_ind)]
                inits['Racket{}.theta'.format(dir_ind)] = fit_results['Stick{}.theta'.format(dir_ind)]
                inits['Racket{}.phi'.format(dir_ind)] = fit_results['Stick{}.phi'.format(dir_ind)]
        elif model_name.startswith('AxCaliber'):
            fit_results = get_model_fit('BallStick_r1')
            inits.update(get_subset(free_parameters, fit_results))
            inits['GDRCylinders.theta'] = fit_results['Stick0.theta']
            inits['GDRCylinders.phi'] = fit_results['Stick0.phi']
        elif model_name.startswith('ActiveAx'):
            fit_results = get_model_fit('BallStick_r1')
            inits.update(get_subset(free_parameters, fit_results))
            inits['w_ic.w'] = fit_results['w_stick0.w'] / 2.0
            inits['w_ec.w'] = fit_results['w_stick0.w'] / 2.0
            inits['w_csf.w'] = fit_results['w_ball.w']
            inits['CylinderGPD.theta'] = fit_results['Stick0.theta']
            inits['CylinderGPD.phi'] = fit_results['Stick0.phi']
        elif model_name.startswith('QMT_ReducedRamani'):
            inits['S0.s0'] = np.mean(input_data.signal4d, axis=-1)

        return inits

    cl_environments = None
    if cl_device_ind is not None:
        cl_environments = get_cl_devices(cl_device_ind)

    with mot_config_context(mot.configuration.RuntimeConfigurationAction(cl_environments=cl_environments)):
        return get_init_data(model_name)


def get_batch_fitting_function(total_nmr_subjects, models_to_fit, output_folder,
                               recalculate=False, cl_device_ind=None, double_precision=False,
                               tmp_results_dir=True, use_gradient_deviations=False):
    """Get the batch fitting function that can fit all desired models on a subject.

    Args:
        total_nmr_subjects (int): the total number of subjects we are fitting.
        models_to_fit (list of str): A list of models to fit to the data.
        output_folder (str): the folder in which to place the output
        recalculate (boolean): If we want to recalculate the results if they are already present.
        cl_device_ind (int): the index of the CL device to use. The index is from the list from the function
            get_cl_devices().
        double_precision (boolean): if we would like to do the calculations in double precision
        tmp_results_dir (str, True or None): The temporary dir for the calculations. Set to a string to use
            that path directly, set to True to use the config value, set to None to disable.
        use_gradient_deviations (boolean): if you want to use the gradient deviations if present
    """
    logger = logging.getLogger(__name__)

    @contextmanager
    def timer(subject_id):
        start_time = timeit.default_timer()
        yield
        logger.info('Fitted all models on subject {0} in time {1} (h:m:s)'.format(
            subject_id, time.strftime('%H:%M:%S', time.gmtime(timeit.default_timer() - start_time))))

    class FitFunc:

        def __init__(self):
            self._index_counter = 0

        def __call__(self, subject_info):
            from mdt import fit_model

            logger.info('Going to process subject {}, ({} of {}, we are at {:.2%})'.format(
                subject_info.subject_id, self._index_counter + 1, total_nmr_subjects,
                self._index_counter / total_nmr_subjects))
            self._index_counter += 1

            output_dir = os.path.join(output_folder, subject_info.subject_id)

            if all(model_output_exists(model, output_dir) for model in models_to_fit) and not recalculate:
                logger.info('Skipping subject {0}, output exists'.format(subject_info.subject_id))
                return

            logger.info('Loading the data (DWI, mask and protocol) of subject {0}'.format(subject_info.subject_id))
            input_data = subject_info.get_input_data(use_gradient_deviations)

            with timer(subject_info.subject_id):
                for model in models_to_fit:
                    if isinstance(model, str):
                        model_name = model
                    else:
                        model_name = model.name

                    logger.info('Going to fit model {0} on subject {1}'.format(model_name, subject_info.subject_id))

                    try:
                        fit_model(model, input_data, output_dir,
                                  recalculate=recalculate,
                                  cl_device_ind=cl_device_ind,
                                  double_precision=double_precision,
                                  tmp_results_dir=tmp_results_dir,
                                  use_cascaded_inits=True)

                    except InsufficientProtocolError as ex:
                        logger.info('Could not fit model {0} on subject {1} '
                                    'due to protocol problems. {2}'.format(model_name, subject_info.subject_id, ex))
                    else:
                        logger.info('Done fitting model {0} on subject {1}'.format(model_name, subject_info.subject_id))

    return FitFunc()


def fit_composite_model(model, input_data, output_folder, method, tmp_results_dir,
                        recalculate=False, optimizer_options=None):
    """Fits the composite model and returns the results as ROI lists per map.

     Args:
        model (:class:`~mdt.models.base.EstimableModel`): An implementation of an composite model
            that contains the model we want to optimize.
        input_data (:class:`~mdt.utils.MRIInputData`): The input data object for the model.
        output_folder (string): The path to the folder where to place the output.
            The resulting maps are placed in a subdirectory (named after the model name) in this output folder.
        method (str): The optimization routine to use.
        tmp_results_dir (str): the main directory to use for the temporary results
        recalculate (boolean): If we want to recalculate the results if they are already present.
        optimizer_options (dict): the additional optimization options
    """
    logger = logging.getLogger(__name__)
    output_path = os.path.join(output_folder, model.name)

    if not model.is_input_data_sufficient(input_data):
        raise InsufficientProtocolError(
            'The given protocol is insufficient for this model. '
            'The reported errors where: {}'.format(model.get_input_data_problems(input_data)))

    if not recalculate and model_output_exists(model, output_folder):
        maps = get_all_nifti_data(output_path)
        logger.info('Not recalculating {} model'.format(model.name))
        return create_roi(maps, input_data.mask)

    with per_model_logging_context(output_path):
        logger.info('Using MDT version {}'.format(__version__))
        logger.info('Preparing for model {0}'.format(model.name))

        model.set_input_data(input_data)

        if recalculate:
            if os.path.exists(output_path):
                list(map(os.remove, glob.glob(os.path.join(output_path, '*.nii*'))))
                if os.path.exists(os.path.join(output_path + 'covariances')):
                    shutil.rmtree(os.path.join(output_path + 'covariances'))

        if not os.path.exists(output_path):
            os.makedirs(output_path)

        with _model_fit_logging(logger, model.name, model.get_free_param_names()):
            tmp_dir = get_full_tmp_results_path(output_path, tmp_results_dir)
            logger.info('Saving temporary results in {}.'.format(tmp_dir))

            worker = FittingProcessor(method, model, input_data.mask,
                                      input_data.nifti_header, output_path,
                                      tmp_dir, recalculate, optimizer_options=optimizer_options)

            processing_strategy = get_processing_strategy('optimization')
            return processing_strategy.process(worker)


@contextmanager
def _model_fit_logging(logger, model_name, free_param_names):
    """Adds logging information around the processing."""
    def calculate_run_days(runtime):
        if runtime > 24 * 60 * 60:
            return int(runtime // (24. * 60 * 60))
        return 0

    minimize_start_time = timeit.default_timer()
    logger.info('Fitting {} model'.format(model_name))
    logger.info('The {} parameters we will fit are: {}'.format(len(free_param_names), free_param_names))
    yield
    run_time = timeit.default_timer() - minimize_start_time
    run_time_str = str(calculate_run_days(run_time)) + ':' + time.strftime('%H:%M:%S', time.gmtime(run_time))
    logger.info('Fitted {0} model with runtime {1} (d:h:m:s).'.format(model_name, run_time_str))
